import dataclasses
from enum import Enum
from pathlib import Path
import platform
from typing import Callable

import cv2
import numpy as np
from ..frame_reader import LocalVideoFileReader, LocalDirReader, FrameReader
from ..recorder import AbstractRecorder, SimpleRecorder


class SupportedOS(Enum):
    LINUX = "Linux"
    WINDOWS = "Windows"


CURRENT_OS = SupportedOS(platform.system())


@dataclasses.dataclass
class KeyFunction:
    key: str
    func: Callable
    description: str


def get_frame_reader(video_source):
    if isinstance(video_source, (str, Path)):
        path = Path(video_source)
        if path.is_file():
            frame_reader = LocalVideoFileReader(video_source)
        elif path.is_dir():
            frame_reader = LocalDirReader
        else:
            raise IOError(f"{video_source} not found")
    elif isinstance(video_source, FrameReader):
        frame_reader = video_source
    else:
        raise ValueError(
            "video_source can needs to be one of {path to a video file, path to frame folder,"
            " an object of a class that implements FrameReader}"
        )
    return frame_reader


def get_recorder(record):
    if isinstance(record, AbstractRecorder):
        recorder = record
    elif isinstance(record, bool):
        recorder = SimpleRecorder() if record else None
    else:
        raise ValueError(
            "record can needs to be one of {True, False," " an object of a class that implements AbstractRecorder}"
        )
    return recorder


def hist_eq_uint16(img):
    hist, bins = np.histogram(img.flatten(), 65536, (0, 65536))  # Collect 16 bits histogram (65536 = 2^16).
    cdf = hist.cumsum()

    cdf_m = np.ma.masked_equal(cdf, 0)  # Find the minimum histogram value (excluding 0)
    cdf_m = (cdf_m - cdf_m.min()) * 65535 / (cdf_m.max() - cdf_m.min())
    cdf = np.ma.filled(cdf_m, 0).astype("uint16")

    # Now we have the look-up table...
    hist_eq_img = cdf[img]
    return hist_eq_img


def get_screen_adjusted_frame_size(screen_size, frame_width, frame_height):
    screen_w, screen_h = screen_size
    if screen_w is None:
        return frame_width, frame_height

    adjusted_w = screen_w
    adjusted_h = screen_h
    if screen_w != frame_width:
        w_ratio = screen_w / frame_width
        adjusted_w = screen_w
        adjusted_h = frame_height * w_ratio
    if screen_h < adjusted_h:
        h_ratio = screen_h / adjusted_h
        adjusted_h = screen_h
        adjusted_w *= h_ratio

    return int(adjusted_w), int(adjusted_h)


def is_window_closed_by_mouse_click(window_name):
    return cv2.getWindowProperty(window_name, cv2.WND_PROP_VISIBLE) < 1



import os

from PIL import Image
import win32gui
import win32con
import win32api
from PIL.Image import Resampling


def set_icon_windows(window_name, icon_path):
    """
    Set the icon for an OpenCV window using a JPG image

    Parameters:
    window_name (str): Name of the OpenCV window
    icon_path (str): Path to the JPG icon image
    """
    try:
        # Convert JPG to ICO format in memory
        img = Image.open(icon_path)
        # Resize to standard icon size if needed
        img = img.resize((32, 32), resample=Resampling.BICUBIC)

        # Create temporary .ico file
        ico_path = "icon.ico"
        img.save(ico_path, format='ICO')

        # Get the window handle
        hwnd = win32gui.FindWindow(None, window_name)

        if hwnd:
            # Set the icon
            icon = win32gui.LoadImage(
                0, ico_path, win32con.IMAGE_ICON,
                0, 0, win32con.LR_LOADFROMFILE
            )
            # Set both small and big icons
            win32api.SendMessage(
                hwnd, win32con.WM_SETICON,
                win32con.ICON_SMALL, icon
            )
            win32api.SendMessage(
                hwnd, win32con.WM_SETICON,
                win32con.ICON_BIG, icon
            )

        os.remove(ico_path)
    except Exception:
        pass


VK_CODE_MAP = {
    0x08: "backspace",
    0x09: "tab",
    0x0C: "clear",
    0x0D: "enter",
    0x10: "shift",
    0x11: "ctrl",
    0x12: "alt",
    0x13: "pause",
    0x14: "caps_lock",
    0x1B: "esc",
    0x20: "spacebar",
    0x21: "page_up",
    0x22: "page_down",
    0x23: "end",
    0x24: "home",
    0x25: "left_arrow",
    0x26: "up_arrow",
    0x27: "right_arrow",
    0x28: "down_arrow",
    0x29: "select",
    0x2A: "print",
    0x2B: "execute",
    0x2C: "print_screen",
    0x2D: "ins",
    0x2E: "del",
    0x2F: "help",
    0x30: "0",
    0x31: "1",
    0x32: "2",
    0x33: "3",
    0x34: "4",
    0x35: "5",
    0x36: "6",
    0x37: "7",
    0x38: "8",
    0x39: "9",
    0x41: "a",
    0x42: "b",
    0x43: "c",
    0x44: "d",
    0x45: "e",
    0x46: "f",
    0x47: "g",
    0x48: "h",
    0x49: "i",
    0x4A: "j",
    0x4B: "k",
    0x4C: "l",
    0x4D: "m",
    0x4E: "n",
    0x4F: "o",
    0x50: "p",
    0x51: "q",
    0x52: "r",
    0x53: "s",
    0x54: "t",
    0x55: "u",
    0x56: "v",
    0x57: "w",
    0x58: "x",
    0x59: "y",
    0x5A: "z",
    0x60: "numpad_0",
    0x61: "numpad_1",
    0x62: "numpad_2",
    0x63: "numpad_3",
    0x64: "numpad_4",
    0x65: "numpad_5",
    0x66: "numpad_6",
    0x67: "numpad_7",
    0x68: "numpad_8",
    0x69: "numpad_9",
    0x6A: "multiply_key",
    0x6B: "add_key",
    0x6C: "separator_key",
    0x6D: "subtract_key",
    0x6E: "decimal_key",
    0x6F: "divide_key",
    0x70: "F1",
    0x71: "F2",
    0x72: "F3",
    0x73: "F4",
    0x74: "F5",
    0x75: "F6",
    0x76: "F7",
    0x77: "F8",
    0x78: "F9",
    0x79: "F10",
    0x7A: "F11",
    0x7B: "F12",
    0x7C: "F13",
    0x7D: "F14",
    0x7E: "F15",
    0x7F: "F16",
    0x80: "F17",
    0x81: "F18",
    0x82: "F19",
    0x83: "F20",
    0x84: "F21",
    0x85: "F22",
    0x86: "F23",
    0x87: "F24",
    0x90: "num_lock",
    0x91: "scroll_lock",
    0xA0: "left_shift",
    0xA1: "right_shift ",
    0xA2: "left_control",
    0xA3: "right_control",
    0xA4: "left_menu",
    0xA5: "right_menu",
    0xA6: "browser_back",
    0xA7: "browser_forward",
    0xA8: "browser_refresh",
    0xA9: "browser_stop",
    0xAA: "browser_search",
    0xAB: "browser_favorites",
    0xAC: "browser_start_and_home",
    0xAD: "volume_mute",
    0xAE: "volume_Down",
    0xAF: "volume_up",
    0xB0: "next_track",
    0xB1: "previous_track",
    0xB2: "stop_media",
    0xB3: "play/pause_media",
    0xB4: "start_mail",
    0xB5: "select_media",
    0xB6: "start_application_1",
    0xB7: "start_application_2",
    0xF6: "attn_key",
    0xF7: "crsel_key",
    0xF8: "exsel_key",
    0xFA: "play_key",
    0xFB: "zoom_key",
    0xFE: "clear_key",
    0xBB: "+",
    0xBC: ",",
    0xBD: "-",
    0xBE: ".",
    0xBF: "/",
    0xC0: "`",
    0xBA: ";",
    0xDB: "[",
    0xDC: "\\",
    0xDD: "]",
    0xDE: "'",
    0xC0: "`",
}



from pathlib import Path
import cv2
import numpy as np
from cvvideoplayer.frame_editors.base_frame_edit_callback import BaseFrameEditCallback
from cvvideoplayer.video_player import VideoPlayer

from typing import Tuple


class FrameSnapshot(BaseFrameEditCallback):
    """
    In charge of saving a specific frame to a file
    """

    def __init__(self, output_video_path: Path, output_shape: Tuple[int, int], frame_num: int):
        super().__init__(enable_by_default=True)
        self._output_video_path = output_video_path
        self._output_shape = output_shape
        self._frame_num = frame_num

    def before_frame_resize(self, video_player: VideoPlayer, frame: np.ndarray, frame_num: int):
        if frame_num == self._frame_num:
            cv2.imwrite(str(self._output_video_path), cv2.resize(frame, self._output_shape))
            print(f"Saved frame {frame_num} to {self._output_video_path}")

            # TODO: This is a hack to make sure we only save the frame once, but it's not a good solution
            self._enabled = False  # For some reason, we get here twice, so we need to disable it after the first time
        return frame



import argparse
from pathlib import Path
from typing import List
import sys
from cvvideoplayer.frame_editors.base_frame_edit_callback import BaseFrameEditCallback
from cvvideoplayer.frame_editors.basic_frame_editors.frame_info_overlay import (
    FrameInfoOverlay,
)
import inspect
from cvvideoplayer.frame_editors.basic_frame_editors.frame_snapshot import FrameSnapshot
from cvvideoplayer.frame_editors.optical_flow_plotter import OpticalFlowPlotter
from cvvideoplayer.utils.ui_utils import InputType, SingleInput
from cvvideoplayer.video_player import VideoPlayer, create_video_player
import cv2


TEST_ASSETS_DIR = Path(__file__).parent / "expected_images"


# python sucks
def plus_plus(integer_in_list_wrapper: List[int]):
    res = integer_in_list_wrapper[0]
    integer_in_list_wrapper[0] += 1
    return res


def create_video_player_for_snapshot(
    source_path: Path,
    output_path,
    frame_num: int,
    frame_edit_callbacks: List[BaseFrameEditCallback] = None,
):
    video_player = create_video_player(
        video_source=Path(source_path),
        frame_edit_callbacks=frame_edit_callbacks + [FrameSnapshot(output_path, (300, 200), frame_num)],
    )

    return video_player


def take_snapshot(video_player: VideoPlayer, key_recordings: List[str]):
    video_player._open_player()
    for key in key_recordings:
        video_player._input_parser._ui_queue.put(SingleInput(InputType.KeyPress, key))

    video_player._run_player_loop()
    cv2.destroyAllWindows()


def compare_images(expected_path: Path, actual_path: Path, diff_path: Path):
    assert actual_path.exists(), f"Something went wrong, the actual image was not saved at {actual_path}"
    # assert that the two images are the same
    expected_frame = cv2.imread(str(expected_path))
    actual_frame = cv2.imread(str(actual_path))
    frame_diff = cv2.absdiff(expected_frame, actual_frame)
    if frame_diff.any():
        cv2.imwrite(str(diff_path), frame_diff)
        return False
    return True


def run_test(
    key_recordings: List[str],
    frame_num,
    record: bool,
    test_name: str,
    frame_edit_callbacks: List[BaseFrameEditCallback] = None,
):
    expected_path = TEST_ASSETS_DIR / f"{test_name}_expected.png"
    actual_path = TEST_ASSETS_DIR / f"{test_name}_actual.png"
    diff_path = TEST_ASSETS_DIR / f"{test_name}_diff.png"
    video_player = create_video_player_for_snapshot(
        source_path=Path(__file__).parent / "../assets/example_video.mp4",
        output_path=expected_path if record else actual_path,
        frame_num=frame_num,
        frame_edit_callbacks=frame_edit_callbacks,
    )

    if record:
        expected_path.unlink(missing_ok=True)
    actual_path.unlink(missing_ok=True)
    diff_path.unlink(missing_ok=True)
    take_snapshot(video_player, key_recordings)

    if not expected_path.exists():
        print(
            f"{test_name}: [ERROR] Expected image not found at {expected_path}, please run the test in record mode first",
            file=sys.stderr,
        )
        sys.exit(1)

    if not record:
        print(f"{test_name}: ", end="")
        res = compare_images(expected_path, actual_path, diff_path)
        if res:
            print("[PASS]")
        else:
            print(
                "[FAIL]",
                f"Expected image and actual image are different, diff image saved at {diff_path}",
            )


def test_simple_image(record=False):
    test_name = inspect.currentframe().f_code.co_name
    frame_edit_callbacks = []

    test_idx = [0]

    frame_num = 3
    key_recordings = ["right" for _ in range(frame_num)] + ["esc"]

    run_test(
        key_recordings,
        frame_num,
        record,
        f"{test_name}_{plus_plus(test_idx)}",
        frame_edit_callbacks,
    )

    frame_num = 10
    key_recordings = ["ctrl+right"] + ["esc"]

    run_test(
        key_recordings,
        frame_num,
        record,
        f"{test_name}_{plus_plus(test_idx)}",
        frame_edit_callbacks,
    )


def test_frame_info_overlay(record=False):
    test_name = inspect.currentframe().f_code.co_name
    frame_edit_callbacks = [FrameInfoOverlay()]

    test_idx = [0]

    frame_num = 15
    key_recordings = ["right" for _ in range(frame_num)] + ["esc"]

    run_test(
        key_recordings,
        frame_num,
        record,
        f"{test_name}_{plus_plus(test_idx)}",
        frame_edit_callbacks,
    )

    frame_num = 7
    key_recordings = ["right" for _ in range(frame_num)] + ["esc"]

    run_test(
        key_recordings,
        frame_num,
        record,
        f"{test_name}_{plus_plus(test_idx)}",
        frame_edit_callbacks,
    )

    frame_num = 3
    key_recordings = ["right" for _ in range(frame_num)] + [
        "ctrl+f",
        "esc",
    ]

    run_test(
        key_recordings,
        frame_num,
        record,
        f"{test_name}_{plus_plus(test_idx)}",
        frame_edit_callbacks,
    )


def test_optical_flow(record=False):
    test_name = inspect.currentframe().f_code.co_name
    frame_edit_callbacks = [OpticalFlowPlotter(enable_by_default=True)]

    test_idx = [0]

    frame_num = 15
    key_recordings = ["right" for _ in range(frame_num)] + [
        "esc",
    ]

    run_test(
        key_recordings,
        frame_num,
        record,
        f"{test_name}_{plus_plus(test_idx)}",
        frame_edit_callbacks,
    )

    frame_num = 7
    key_recordings = ["right" for _ in range(frame_num)] + ["esc"]

    run_test(
        key_recordings,
        frame_num,
        record,
        f"{test_name}_{plus_plus(test_idx)}",
        frame_edit_callbacks,
    )

    frame_num = 3
    key_recordings = ["right" for _ in range(frame_num)] + [
        "esc",
    ]

    run_test(
        key_recordings,
        frame_num,
        record,
        f"{test_name}_{plus_plus(test_idx)}",
        frame_edit_callbacks,
    )


def run(record=False):
    TEST_ASSETS_DIR.mkdir(exist_ok=True)
    test_simple_image(record)
    test_frame_info_overlay(record)
    test_optical_flow(record)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="record expected images from the player and replay against a new implementation to show differences in the output",
        allow_abbrev=False,
    )
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("--record", action="store_true", default=False, help="Record expected images")
    group.add_argument("--replay", action="store_true", default=False, help="Replay and run the tests")

    args = parser.parse_args()

    run(args.record and not args.replay)



from dataclasses import dataclass
from enum import Enum
from typing import Tuple, Optional


class BboxFormat(Enum):
    xywh = "xywh"  # x_min, y_min, width, height
    xyxy = "xyxy"  # x_min, y_min, x_max, y_max
    xcycwh = "xcycwh"  # x_center, y_center, width, height


@dataclass
class Bbox:
    x1: int
    y1: int
    width: int
    height: int
    color: Optional[Tuple[int, int, int]] = None
    above_label: Optional[str] = None
    below_label: Optional[str] = None

    @classmethod
    def init_with_xyxy(cls, x1, y1, x2, y2, color=None, above_label=None, below_label=None):
        return cls(x1, y1, x2 - x1, y2 - y1, color, above_label, below_label)

    @property
    def x2(self):
        return self.x1 + self.width

    @property
    def y2(self):
        return self.y1 + self.height

    @property
    def x_center(self):
        return self.x1 + (self.width / 2)

    @property
    def y_center(self):
        return self.y1 + (self.height / 2)

    @property
    def area(self):
        return self.width * self.height

    def get_normalized_bbox(
            self,
            frame_width: int,
            frame_height: int,
            bbox_format: BboxFormat = BboxFormat.xywh
    ) -> Tuple[float, float, float, float]:
        frame_width = float(frame_width)
        frame_height = float(frame_height)
        if bbox_format == BboxFormat.xywh:
            return self.x1 / frame_width, self.y1 / frame_height, self.width / frame_width, self.height / frame_height
        if bbox_format == BboxFormat.xyxy:
            return self.x1 / frame_width, self.y1 / frame_height, self.x2 / frame_width, self.y2 / frame_height
        if bbox_format == BboxFormat.xcycwh:
            return self.x_center / frame_width, self.y_center / frame_height, self.width / frame_width, self.height / frame_height



import os
import sys
from pathlib import Path


def change_cwd_to_tests_dir():
    os.chdir(str(Path(__file__).absolute().parent))


def add_project_root_to_path():
    sys.path.append(str(Path(__file__).absolute().parent.parent))



from .frame_reader import *
from .recorder import SimpleRecorder, AbstractRecorder
from .video_player import create_video_player
from .video_player import DeprecatedVideoPlayer as VideoPlayer
from .utils.bbox_utils import Bbox
from .utils.video_player_utils import KeyFunction



from collections import defaultdict
from typing import Dict

import cv2
from ..utils.video_player_utils import KeyFunction
from ..utils.ui_utils import SingleInput, InputType


class InputHandler:
    def __init__(self, window_name: str):
        self._keymap: Dict[str, KeyFunction] = {}
        self._keymap_description: Dict[str, list] = defaultdict(list)
        self._window_name = window_name

    def register_key_function(self, key_function: KeyFunction, callback_name: str) -> None:
        key = key_function.key
        if "+" in key:
            modifiers = sorted([modifier for modifier in key.split("+")[:-1] if modifier])
            key_without_modifiers = key.split("+")[-1] or "+"
            key = "+".join(modifiers + [key_without_modifiers])
        else:
            key_without_modifiers = key
            modifiers = []

        if key_without_modifiers.isnumeric():
            general_num_key = "+".join(modifiers + ["num"])
            if general_num_key in self._keymap:
                raise KeyError(f"{general_num_key} is registered: {self._keymap[general_num_key].description}")

        if key in self._keymap:
            raise KeyError(
                f"{key} is already registered (Registered action description: {self._keymap[key].description})"
            )

        self._keymap[key] = key_function
        if key_function.description:
            self._keymap_description[callback_name].append(f"{key:20.20}: {key_function.description:60.60}")

    def get_keymap_description(self) -> dict:
        return self._keymap_description

    def unregister_key_function(self, key: str) -> None:
        if key in self._keymap:
            del self._keymap[key]

    def handle_input(self, single_input: SingleInput) -> None:
        if single_input.input_type == InputType.KeyPress:
            self._handle_keypress(key_str=single_input.input_data)
        elif single_input.input_type == InputType.MouseScroll:
            self._handle_mouse_scroll(*single_input.input_data)

    def _handle_keypress(self, key_str):
        key_without_modifiers = key_str.split("+")[-1]
        if key_without_modifiers.isnumeric():
            general_num_key = key_str.replace(key_without_modifiers, "num")
        else:
            general_num_key = None

        if general_num_key in self._keymap:
            self._keymap[general_num_key].func(key_without_modifiers)
        elif key_str in self._keymap:
            self._keymap[key_str].func()
        else:
            print(f"{key_str} is not registered in the keymap")

    def _handle_mouse_scroll(self, x, y, dx, dy):
        if "mouse_scroll" not in self._keymap:
            return
        win_x, win_y, win_w, win_h = cv2.getWindowImageRect(self._window_name)
        curser_x = x - win_x
        curser_y = y - win_y
        if curser_x < 0 or curser_y < 0:
            return
        self._keymap["mouse_scroll"].func(curser_x, curser_y, dy)



import abc
import ctypes
import subprocess
import time
from pathlib import Path
from queue import Empty
from typing import Optional, List, Union, Tuple
from functools import partial

import cv2
import numpy as np

from .frame_editors import BaseFrameEditCallback
from .frame_reader import FrameReader
from .input_management.base_input_parser import BaseInputParser
from .input_management.input_handler import InputHandler
from .recorder import AbstractRecorder
from .utils.video_player_utils import (
    get_screen_adjusted_frame_size,
    KeyFunction,
    is_window_closed_by_mouse_click,
    get_frame_reader,
    get_recorder,
    CURRENT_OS,
    SupportedOS,
)

if CURRENT_OS == SupportedOS.LINUX:
    import Xlib
    from .utils.linux_os_utils import set_icon_linux
    from .input_management.linux_input_parser import LinuxInputParser
elif CURRENT_OS == SupportedOS.WINDOWS:
    from .utils.windows_os_utils import set_icon_windows
    from .input_management.windows_input_parser import WindowsInputParser


class VideoPlayer(abc.ABC):
    def __init__(
        self,
        video_source: Union[str, Path, FrameReader],
        start_from_frame: int = 0,
        frame_edit_callbacks: Optional[List[BaseFrameEditCallback]] = None,
        record: Union[bool, AbstractRecorder] = False,
    ):
        """
        Params:
        - video_source : Union[str, Path, FrameReader] The source of the video to be played. It can be a file path, a
         directory path, or a FrameReader instance.
        - start_from_frame : int, optional The frame number to start the video from (default is 0).
        - frame_edit_callbacks : list, optional A list of frame editing callbacks (default is None).
         Each callback must be an instance of BaseFrameEditCallback.
        - record : Union[bool, AbstractRecorder], optional Whether to record the video or not (default is False).
        It can also be an instance of AbstractRecorder for custom recording functionality.
        """
        self._window_name = "CVvideoPlayer"
        self.frame_reader = get_frame_reader(video_source)
        self.input_handler = InputHandler(self._window_name)
        self._recorder = get_recorder(record)

        self._last_frame = len(self.frame_reader) - 1
        self._current_frame_num = start_from_frame

        self._screen_size = self._get_screen_size()
        self._screen_adjusted_frame_size = None
        self._original_frame_size = None

        self._frame_edit_callbacks = frame_edit_callbacks or []

        self._play = False
        self._exit = False
        self._add_default_key_functions()
        self._setup_callbacks()

        current_frame = self._get_current_frame()
        self._screen_adjusted_frame_size = get_screen_adjusted_frame_size(
            screen_size=self._screen_size,
            frame_width=current_frame.shape[1],
            frame_height=current_frame.shape[0],
        )

    def __enter__(self):
        return self

    def __exit__(self, *args):
        cv2.destroyAllWindows()
        if self._recorder is not None:
            self._recorder.teardown()
        for callback in self._frame_edit_callbacks:
            callback.teardown()
        self._input_parser.stop()

    def run(self) -> None:
        try:
            self._open_player()
            self._input_parser.start()
            self._run_player_loop()
        finally:
            self.__exit__()

    @abc.abstractmethod
    def _get_screen_size(self) -> Tuple[int, int]:
        """
        implemented per platform
        """

    @abc.abstractmethod
    def _get_in_focus_window_id(self) -> int:
        """
        implemented per platform
        """

    @property
    @abc.abstractmethod
    def _input_parser(self) -> BaseInputParser:
        """
        implemented per platform
        """

    @abc.abstractmethod
    def _set_icon(self):
        """
        implemented per platform
        """

    def _run_player_loop(self):
        while not self._exit:
            if self._get_in_focus_window_id() != self._window_id:
                self._input_parser.pause()
                if is_window_closed_by_mouse_click(window_name=self._window_name):
                    break
                else:
                    cv2.waitKey(100)
                    continue
            self._input_parser.resume()
            try:
                single_input = self._input_parser.get_input()
            except Empty:
                cv2.waitKey(1)
                continue

            self.input_handler.handle_input(single_input)

            if self._play:
                self._play_continuously()
            else:
                frame_for_display = self._create_frame_to_display()
                self._show_frame(frame_for_display)

    def _open_player(self) -> None:
        frame_for_display = self._create_frame_to_display()
        self._show_frame(frame_for_display)
        time.sleep(0.5)  # make sure video player is up before checking the window id
        self._window_id = self._get_in_focus_window_id()
        self._set_icon()
        cv2.waitKey(50)

    def _setup_callbacks(self):
        for callback in self._frame_edit_callbacks:
            assert isinstance(callback, BaseFrameEditCallback), (
                "frame_editor must be a derived class of" " BaseFrameEditor"
            )
            for key_function in callback.key_function_to_register:
                assert isinstance(key_function, KeyFunction), (
                    f"{callback.__class__.__name__} is trying to register a key function"
                    f" (key = {key_function.key}) which is not an instance of KeyFunction"
                )
                self.input_handler.register_key_function(
                    key_function,
                    callback.__class__.__name__,
                )

            callback.setup(self, frame=self._get_current_frame())

    def _create_frame_to_display(self) -> np.ndarray:
        frame = self._get_current_frame().copy()

        for callback in self._frame_edit_callbacks:
            if not callback.enabled:
                continue
            shape_before_edit = frame.shape[:2]
            frame = callback.before_frame_resize(self, frame, self._current_frame_num)
            assert frame.shape[:2] == shape_before_edit, "callbacks can not alter the frame's shape before resize"

        frame = self._crop_and_resize_frame(frame)

        for callback in self._frame_edit_callbacks:
            if not callback.enabled:
                continue
            frame = callback.after_frame_resize(self, frame, self._current_frame_num)

        #  we resize again in case frame size has been changed in "after_frame_resize" hooks
        if (
            frame.shape[0] != self._screen_adjusted_frame_size[1]
            or frame.shape[1] != self._screen_adjusted_frame_size[0]
        ):
            frame = cv2.resize(frame, self._screen_adjusted_frame_size)

        if self._recorder is not None:
            self._recorder.write_frame_to_video(self, frame, self._current_frame_num)

        return frame

    def _show_frame(self, frame):
        cv2.imshow(winname=self._window_name, mat=frame)
        cv2.waitKey(1)

    def _play_continuously(self) -> None:
        while (not self._input_parser.has_input()) and self._play and not self._exit:
            self._change_current_frame_num(change_by=1)
            frame = self._create_frame_to_display()
            self._show_frame(frame)

    def _crop_and_resize_frame(self, frame) -> np.ndarray:
        frame = cv2.resize(frame, self._screen_adjusted_frame_size)
        return frame

    def _change_current_frame_num(self, change_by: int) -> None:
        if change_by > 0 and self._current_frame_num == self._last_frame:
            self._play = False
            return

        if change_by < 0 and self._current_frame_num == 0:
            return

        self._current_frame_num = max(0, min(self._current_frame_num + change_by, self._last_frame))

    def _get_current_frame(self) -> np.ndarray:
        return self.frame_reader.get_frame(self._current_frame_num)

    def _pause_and_change_current_frame(self, change_by: int) -> None:
        self._play = False
        self._change_current_frame_num(change_by)

    def _play_pause(self):
        self._play = not self._play

    def _set_exit_to_true(self):
        self._exit = True

    def _add_default_key_functions(self) -> None:
        default_key_functions = [
            KeyFunction("space", self._play_pause, "Play/Pause video"),
            KeyFunction("right", partial(self._pause_and_change_current_frame, 1), "Next frame"),
            KeyFunction("left", partial(self._pause_and_change_current_frame, -1), "Previous frame"),
            KeyFunction("ctrl+right", partial(self._pause_and_change_current_frame, 10), "10 frames forward"),
            KeyFunction("ctrl+left", partial(self._pause_and_change_current_frame, -10), "10 frames back"),
            KeyFunction("ctrl+shift+right", partial(self._pause_and_change_current_frame, 50), "50 frames forward"),
            KeyFunction("ctrl+shift+left", partial(self._pause_and_change_current_frame, -50), "50 frames back"),
            KeyFunction("esc", self._set_exit_to_true, "Exit gracefully"),
        ]

        for key_function in default_key_functions:
            self.input_handler.register_key_function(key_function, "Video Control")


class WindowsVideoPlayer(VideoPlayer):
    def __init__(self, **video_player_kwargs):
        ctypes.windll.shcore.SetProcessDpiAwareness(2)
        super().__init__(**video_player_kwargs)
        self._zoom_factor = 1.0
        self._zoom_crop_xywh = None
        current_frame = self._get_current_frame()
        self._original_frame_size = tuple(current_frame.shape[:2][::-1])
        self._zoom_crop_xywh = (
            0,
            0,
            self._original_frame_size[0],
            self._original_frame_size[1],
        )

    def _show_frame(self, frame) -> None:
        super()._show_frame(frame)
        cv2.waitKey(1)  # for some reason Windows OS requires an additional waitKey to work properly

    @property
    def _input_parser(self):
        return WindowsInputParser()

    def _get_screen_size(self):
        user32 = ctypes.windll.user32
        screensize = 0.9 * user32.GetSystemMetrics(0), 0.9 * user32.GetSystemMetrics(1)
        return screensize

    def _get_in_focus_window_id(self):
        h_wnd = ctypes.windll.user32.GetForegroundWindow()
        pid = ctypes.wintypes.DWORD()
        ctypes.windll.user32.GetWindowThreadProcessId(h_wnd, ctypes.byref(pid))
        return pid.value

    def _add_default_key_functions(self) -> None:
        super()._add_default_key_functions()
        scroll_to_zoom = KeyFunction(key="mouse_scroll", func=self._set_zoom, description="zoom in and out")
        self.input_handler.register_key_function(key_function=scroll_to_zoom, callback_name="Video Control")

    def _set_zoom(self, curser_x: int, curser_y: int, scroll_direction: int):
        new_zoom_factor = max(1.0, min(5.0, self._zoom_factor + 0.5 * scroll_direction))
        if new_zoom_factor != self._zoom_factor:
            self._zoom_factor = new_zoom_factor
            adjusted_width, adjusted_height = self._screen_adjusted_frame_size
            norm_curser_position = (
                curser_x / adjusted_width,
                curser_y / adjusted_height,
            )
            w = h = 1 / self._zoom_factor
            x_min = norm_curser_position[0] * (1 - w)
            y_min = norm_curser_position[1] * (1 - h)
            self._zoom_crop_xywh = (
                int(self._original_frame_size[0] * x_min),
                int(self._original_frame_size[1] * y_min),
                int(self._original_frame_size[0] * w),
                int(self._original_frame_size[1] * h),
            )

    def _crop_and_resize_frame(self, frame) -> np.ndarray:
        frame = frame[
            self._zoom_crop_xywh[1] : self._zoom_crop_xywh[1] + self._zoom_crop_xywh[3],
            self._zoom_crop_xywh[0] : self._zoom_crop_xywh[0] + self._zoom_crop_xywh[2],
        ]
        frame = cv2.resize(frame, self._screen_adjusted_frame_size)
        return frame

    def _set_icon(self):
        set_icon_windows(self._window_name, icon_path=Path(__file__).parent / "icon.png")


class LinuxVideoPlayer(VideoPlayer):
    def __init__(self, **video_player_kwargs):
        super().__init__(**video_player_kwargs)
        self._display = Xlib.display.Display()

    @property
    def _input_parser(self):
        return LinuxInputParser()

    def _get_in_focus_window_id(self):
        window = self._display.get_input_focus().focus
        if isinstance(window, int):
            return ""
        return window.id

    def _get_screen_size(self):
        screen_size_str = (
            subprocess.check_output('xrandr | grep "\*" | cut -d" " -f4', shell=True).decode().strip().split("\n")[0]
        )
        screen_w, screen_h = screen_size_str.split("x")
        screen_w = 0.85 * int(screen_w)
        screen_h = 0.85 * int(screen_h)
        return screen_w, screen_h

    def _set_icon(self):
        set_icon_linux(self._window_id, self._display, icon_path=str(Path(__file__).parent / "icon.png"))


def create_video_player(**video_kwargs) -> VideoPlayer:
    if CURRENT_OS == SupportedOS.WINDOWS:
        return WindowsVideoPlayer(**video_kwargs)
    elif CURRENT_OS == SupportedOS.LINUX:
        return LinuxVideoPlayer(**video_kwargs)


class DeprecatedVideoPlayer:
    def __init__(
        self,
        video_source: Union[str, Path, FrameReader],
        start_from_frame: int = 0,
        frame_edit_callbacks: Optional[List[BaseFrameEditCallback]] = None,
        record: Union[bool, AbstractRecorder] = False,
    ):
        self._video_player = create_video_player(
            video_source=video_source,
            start_from_frame=start_from_frame,
            frame_edit_callbacks=frame_edit_callbacks,
            record=record,
        )

    def run(self) -> None:
        self._video_player.run()



from .base_input_parser import BaseInputParser
from ..utils.windows_os_utils import VK_CODE_MAP


class WindowsInputParser(BaseInputParser):
    @staticmethod
    def _vk_code_mapper(vk_code):
        return VK_CODE_MAP[vk_code]



from .video_player_utils import CURRENT_OS, SupportedOS

if CURRENT_OS == SupportedOS.LINUX:
    import cv2
    from PIL import Image
    import Xlib.display
    from Xlib import X
    import os


def rgba_to_uint32(r, g, b, a):
    """Convert RGBA values to 32-bit unsigned integer in BGRA format"""
    return (a << 24) | (r << 16) | (g << 8) | b


def set_icon_linux(window_id, display, icon_path):
    try:
        # Get the window handle using Xlib
        # Load and prepare the icon
        icon = Image.open(icon_path)
        icon = icon.convert("RGBA")
        icon = icon.resize((32, 32))
        width, height = icon.size

        # Convert icon to X format
        icon_data = []
        icon_data.extend([width, height])

        for y in range(height):
            for x in range(width):
                r, g, b, a = icon.getpixel((x, y))
                icon_data.append(rgba_to_uint32(r, g, b, a))

        # Get the _NET_WM_ICON atom
        _NET_WM_ICON = display.intern_atom('_NET_WM_ICON')

        # Convert the list to a format Xlib can handle
        data = (height * width + 2) * [0]
        for i in range(len(icon_data)):
            data[i] = icon_data[i]

        # Set the property
        window = display.create_resource_object('window', window_id)
        window.change_property(_NET_WM_ICON, Xlib.Xatom.CARDINAL, 32, data)

        # Flush the changes
        display.flush()
        display.sync()
    except Exception:
        return



import abc
from queue import Queue
from typing import Union

from pynput import keyboard, mouse
from pynput.mouse import Button

from ..utils.ui_utils import Singleton, SingleInput, mouse_button_parser, InputType, MODIFIERS


class BaseInputParser(metaclass=Singleton):
    def __init__(self, allow_queue_buildup=False):
        self._allow_queue_buildup = allow_queue_buildup
        self._ui_queue: Queue[SingleInput] = Queue()
        self._modifiers = set()
        self._listeners = []
        self._queue_is_blocked = False

    def pause(self):
        self._queue_is_blocked = True

    def resume(self):
        self._queue_is_blocked = False

    def has_input(self) -> bool:
        return not self._ui_queue.empty()

    def get_input(self) -> SingleInput:
        return self._ui_queue.get(timeout=0.1)

    def start(self) -> None:
        self._listeners.append(
            keyboard.Listener(
                on_press=self._add_key_press_to_queue,
                on_release=self._add_key_release_to_queue,
            )
        )
        self._listeners.append(
            mouse.Listener(
                on_click=self._add_mouse_click_to_queue,
                on_scroll=self._add_mouse_scroll_to_queue,
            )
        )

        for listener in self._listeners:
            listener.start()

    def stop(self) -> None:
        self._listeners.clear()

    @staticmethod
    @abc.abstractmethod
    def _vk_code_mapper(vk_code) -> str:
        pass

    def _queue_is_open_for_business(self) -> bool:
        return not self._queue_is_blocked and (not self.has_input() or self._allow_queue_buildup)

    def _add_mouse_click_to_queue(self, x: int, y: int, button: Button, is_clicked: bool) -> None:
        mouse_button = mouse_button_parser[button]
        if self._queue_is_open_for_business():
            if is_clicked:
                self._ui_queue.put(SingleInput(InputType.MouseClick, (x, y, mouse_button)))
            # else:
            #     self._ui_queue.put(SingleInput(InputType.MouseRelease, (x, y, mouse_button)))

    def _add_mouse_scroll_to_queue(self, x, y, dx, dy):
        if self._queue_is_open_for_business():
            self._ui_queue.put(SingleInput(InputType.MouseScroll, (x, y, dx, dy)))

    def _add_key_press_to_queue(self, key: keyboard.Key) -> None:
        key_str = self._parse_pynput_key(key)
        if key_str in MODIFIERS:
            self._modifiers.add(key_str)
        else:
            key_str = "+".join(sorted(self._modifiers) + [key_str])
            # To avoid a situation of execution build up due to slow execution time
            if self._queue_is_open_for_business():
                self._ui_queue.put(SingleInput(InputType.KeyPress, key_str))

    def _add_key_release_to_queue(self, key: Union[keyboard.Key, str]) -> None:
        key_str = self._parse_pynput_key(key)
        if key_str in MODIFIERS:
            self._modifiers.discard(key_str)

    def _parse_pynput_key(self, key: Union[keyboard.Key, keyboard.KeyCode]) -> str:
        if str(key) == "<65437>":  # work around for a bug in pynput model that does not convert 5 for some reason
            key = "5"

        if hasattr(key, "vk") and key.vk is not None:
            key_str = self._vk_code_mapper(key.vk).lower()
        else:
            key_str = str(key).replace("'", "")
            key_str = key_str.replace("Key.", "")
            key_str = key_str.split("_", maxsplit=1)[0].lower()

        return key_str



import abc
from dataclasses import dataclass
from enum import Enum, auto
from typing import Tuple, Union

from pynput import mouse


mouse_button_parser = {
    mouse.Button.left: "mouse_left",
    mouse.Button.right: "mouse_right",
    mouse.Button.middle: "mouse_middle",
}

MODIFIERS = {
    "ctrl",
    "alt",
    "shift",
}


class Singleton(abc.ABCMeta):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class InputType(Enum):
    KeyPress = auto()
    KeyRelease = auto()
    MouseScroll = auto()
    MouseClick = auto()
    MouseRelease = auto()
    MouseMove = auto()


@dataclass
class SingleInput:
    input_type: InputType
    input_data: Union[Tuple, str]




from setuptools import setup, find_packages

with open("README.md", 'r') as f:
    long_description = f.read()

setup(
    name="cvvideoplayer",
    version="1.1.5",
    author="Daniel Tomer",
    author_email="danieltomer1@gmail.com",
    description="moduler multi purpose video player for debugging algorithms in python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
    install_requires=[
        "numpy",
        "opencv-python",
        "pynput",
        "matplotlib",
        "python-xlib;platform_system=='Linux'",
        "pywin32;platform_system=='Windows'"
    ],
    python_requires=">=3.8",
    project_urls={
        "Homepage": "https://github.com/danieltomer1/CVVideoPlayer",
    },
    keywords=[
        'opencv',
        'video',
        'player',
        'video player',
        'cvvideoplayer',
        'computer vision',
        'image processing',
        'customizable'
    ]
)
